"""
Support for dealing with (genome scale) sequence alignments. See `core` for
the abstract alignment classes and `maf`, `axt`, and `lav` for readers and
writers in various formats.
"""

from bx.align.core import *  # noqa
